DATABASES = {'default': {
  'ENGINE': 'django.db.backends.sqlite3',
  'NAME': '.testi.sqlite',
}}
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'
INSTALLED_APPS = ['lume', 'testit']
MIDDLEWARE = []
SECRET_KEY = 'epäjärjestelmällistyttämättömyydellänsäkäänköhän'
USE_TZ = True



from decimal import Decimal
from unittest import expectedFailure

from django.db import models
from django.forms import modelform_factory
from django import test

from .mallit import (
  Asiakas,
  Lasku,
  Osoite,
  Paamies,
  Rivi,
)


class Lume(test.TestCase):
  # pylint: disable=invalid-name, unused-variable

  @classmethod
  def setUpTestData(cls):
    super().setUpTestData()
    asiakas = Asiakas.objects.create(nimi='Asiakas')
    paamies = Paamies.objects.create(nimi='Velkoja')
    Osoite.objects.create(asiakas=asiakas, osoite='Katu 1')
    Osoite.objects.create(asiakas=asiakas, osoite='Katu 123 B 4')
    Osoite.objects.create(asiakas=asiakas, osoite='Katu 456')
    lasku_1 = Lasku.objects.create(asiakas=asiakas, paamies=paamies, numero=1)
    lasku_3 = Lasku.objects.create(asiakas=asiakas, paamies=paamies, numero=3)
    lasku_2 = Lasku.objects.create(asiakas=asiakas, paamies=paamies, numero=2)
    Rivi.objects.create(lasku=lasku_3, summa=123, selite='Samppanjaa')
    Rivi.objects.create(lasku=lasku_3, summa=456, selite='Kaviaaria')
    Rivi.objects.create(lasku=lasku_2, summa=789, selite='Jalokiviä')

  def testaa_paikallinen_laskenta(self):
    asiakas = Asiakas.objects.only('pk').get()
    self.assertEqual(asiakas.pisin_osoite.osoite, 'Katu 123 B 4')
    self.assertTrue(asiakas.useita_laskuja)
    # def testaa_only

  def testaa_refresh_from_db(self):
    asiakas = Asiakas.objects.only('pk').get()
    #asiakas.refresh_from_db('pk')
    print(asiakas.__dict__, asiakas._state.__dict__)
    #raise RuntimeError
    self.assertTrue(asiakas.useita_laskuja)
    # def testaa_refresh_from_db

  def testaa_1x_lume(self):
    '''
    Asiakkaan pisin osoite = Katu 123 B 4?
    '''
    self.assertEqual(
      list(Asiakas.objects.values('pisin_osoite__osoite')),
      [{'pisin_osoite__osoite': 'Katu 123 B 4'}],
    )
    # def testaa_1x_lume

  def testaa_2x_lume(self):
    '''
    Viimeisimmän laskun arvokkaimman rivin selite = Kaviaaria?
    '''
    self.assertEqual(
      list(Asiakas.objects.values('viimeisin_lasku__arvokkain_rivi__selite')),
      [{'viimeisin_lasku__arvokkain_rivi__selite': 'Kaviaaria'}],
    )
    # def testaa_2x_lume

  def testaa_lume_laskenta(self):
    '''
    Viimeisimmän laskun arvokkaimman rivin osuus koko summasta on oikein?
    '''
    self.assertEqual(
      Asiakas.objects.get().viimeisin_lasku.arvokkain_osuus,
      Decimal(456) // Decimal(123+ 456),
    )
    self.assertTrue(
      Asiakas.objects.get().viimeisin_lasku.arvo_yli_500,
    )
    self.assertFalse(
      Asiakas.objects.get().vanhin_lasku.arvo_yli_500,
    )
    # def testaa_lume_laskenta

  @expectedFailure
  def testaa_outerref(self):
    '''
    Niiden laskujen lukumäärä, joiden koko summa on yhdellä rivillä = 1?

    – Lumekenttä OuterRef-viittauksen takaa ei toistaiseksi toimi.
    '''
    self.assertEqual(Lasku.objects.filter(
      models.Exists(
        Rivi.objects.filter(
          lasku=models.OuterRef('pk'),
          summa=models.OuterRef('rivien_summa'),
        )
      )
    ).count(), 1)
    # def testaa_outerref_lume

  def testaa_summa(self):
    '''
    Päämiehen laskujen summa = 1368?
    '''
    self.assertEqual(
      list(Paamies.objects.values('laskujen_summa')),
      [{'laskujen_summa': Decimal('1368')}],
    )
    # def testaa_summa

  def testaa_tallennus(self):
    ''' Toimiiko tallennus lumekenttien estämättä? '''
    asiakas = Asiakas(nimi='Asiakas')
    asiakas.save()
    asiakas.refresh_from_db()
    self.assertEqual(asiakas.viimeisin_lasku, None)
    # def testaa_tallennus

  def testaa_luonti_lomakkeen_kautta(self):
    ''' Toimiiko mallipohjaisen lomakkeen luonti ja tallennus? '''
    asiakas = modelform_factory(Asiakas, fields=['nimi'])(
      data={'nimi': 'Uusi Asiakas'},
    ).save()
    asiakas.refresh_from_db
    # def testaa_luonti_lomakkeen_kautta

  def testaa_paivitys_lomakkeen_kautta(self):
    ''' Toimiiko mallipohjaisen lomakkeen luonti ja tallennus? '''
    asiakas = Asiakas.objects.first()
    modelform_factory(Asiakas, fields=['nimi'])(
      data={'nimi': 'Nimi Vaihdettu'},
      instance=asiakas,
    ).save()
    # def testaa_paivitys_lomakkeen_kautta

  def testaa_valmiiksi_laskettu(self):
    self.assertEqual(
      list(Asiakas.objects.lume('viimeisin_lasku').values_list(
        'viimeisin_lasku__arvokkain_rivi__selite'
      )),
      [('Kaviaaria', ), ]
    )
    # def testaa_valmiiksi_laskettu

  # class Lume



# -*- coding: utf-8 -*-

import inspect

from django.db import models

from .kentta import EI_ASETETTU, Lumekentta
from . import puukko


# Nimetään eri kenttäluokat tässä, jotta koodintarkistusalgoritmit eivät
# valita niistä.
# Luettelo per Django 5.1.
AutoField: type
BigAutoField: type
BigIntegerField: type
BinaryField: type
BooleanField: type
CharField: type
CommaSeparatedIntegerField: type
DateField: type
DateTimeField: type
DecimalField: type
DurationField: type
EmailField: type
Field: type
FilePathField: type
FloatField: type
GeneratedField: type
GenericIPAddressField: type
IPAddressField: type
IntegerField: type
NullBooleanField: type
PositiveBigIntegerField: type
PositiveIntegerField: type
PositiveSmallIntegerField: type
SlugField: type
SmallAutoField: type
SmallIntegerField: type
TextField: type
TimeField: type
URLField: type
UUIDField: type
FileField: type
ImageField: type
JSONField: type
ForeignKey: type
OneToOneField: type


# Periytä lumeversio kustakin Djangon kenttätyypistä.
for nimi, luokka in inspect.getmembers(
  models, lambda x: inspect.isclass(x) and issubclass(x, models.Field)
):
  globals()[nimi] = type(nimi, (Lumekentta, luokka), {})
del inspect
del models



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name, protected-access, unused-argument

from contextlib import contextmanager
import functools
import itertools

from django.db.migrations import autodetector
from django.db import models
from django.db.models.options import Options

from .kentta import Lumekentta


def puukota(moduuli, koriste=None, kopioi=None):
  '''
  Korvaa moduulissa olevan metodin tai lisää uuden (`kopioi`).
  '''
  def puukko(funktio):
    toteutus = getattr(moduuli, kopioi or funktio.__name__)
    def uusi_toteutus(*args, **kwargs):
      return funktio(toteutus, *args, **kwargs)
    setattr(
      moduuli, funktio.__name__,
      (koriste or functools.wraps(toteutus))(uusi_toteutus)
    )
  return puukko
  # def puukota


@puukota(autodetector.MigrationAutodetector)
def __init__(oletus, self, *args, **kwargs):
  '''
  Poista lumekentät migraatioiden luonnin yhteydessä
  sekä vanhojen että uusien kenttien listalta.
  '''
  oletus(self, *args, **kwargs)
  for malli in self.from_state.models.values():
    malli.fields = {
      l: f for l, f in malli.fields.items() if not isinstance(f, Lumekentta)
    }
  for malli in self.to_state.models.values():
    malli.fields = {
      l: f for l, f in malli.fields.items() if not isinstance(f, Lumekentta)
    }
  # def __init__


@puukota(Options, koriste=property)
def concrete_fields(oletus, self):
  '''
  Järjestä lumekentät viimeisiksi.

  Tätä tarvitaan uutta riviä luotaessa, jotta todellisten
  sarakkeiden arvot ovat käytettävissä lumekenttiä asetettaessa.
  '''
  return models.options.make_immutable_fields_list(
    "concrete_fields", itertools.chain((
      f for f in self.fields
      if f.concrete and not isinstance(f, Lumekentta)
    ), (
      f for f in self.fields
      if f.concrete and isinstance(f, Lumekentta)
    ))
  )
  # def concrete_fields


@puukota(Options, koriste=property)
def local_concrete_fields(oletus, self):
  '''
  Ohita lumekentät mallin konkreettisia kenttiä kysyttäessä.
  '''
  return models.options.make_immutable_fields_list(
    "local_concrete_fields", (
      f for f in self.local_fields
      if f.concrete and not isinstance(f, Lumekentta)
    )
  )
  # def local_concrete_fields


@puukota(models.query.QuerySet)
def _insert(oletus, self, objs, fields, **kwargs):
  '''
  Poista mahdolliset lumekentät tallennettavista kentistä.
  '''
  return oletus(self, objs, fields=[
    f for f in fields if not isinstance(f, Lumekentta)
  ], **kwargs)
  # def _insert


@puukota(models.query.QuerySet, kopioi='only')
def lume(only, self, *fields):
  '''
  Lisää annetut lumekentät pyydettyjen kenttien listalle, tai tyhjennä lista.
  '''
  if self._fields is not None:
    raise TypeError(
      'Ei voida kutsua metodia .lume()'
      ' aiemman .values()- tai .values_list()-kutsun jälkeen.'
    )
  clone = self._chain()
  if fields == (None,):
    clone.query.tyhjenna_lumekentat()
  else:
    clone.query.lisaa_lumekentat(fields)
  return clone
  # def lume


# Metodia `models.Manager._get_queryset_methods()` on tässä vaiheessa
# jo kutsuttu, joten kopioidaan `lume`-metodi käsin `Manager`-luokkaan:
def m_lume(self, *args, **kwargs):
  return getattr(self.get_queryset(), 'lume')(*args, **kwargs)
models.Manager.lume = m_lume


@puukota(models.sql.query.Query, kopioi='clear_deferred_loading')
def tyhjenna_lumekentat(oletus, self):
  self.pyydetyt_lumekentat = frozenset()
  # def tyhjenna_lumekentat


@puukota(models.sql.query.Query, kopioi='add_deferred_loading')
def lisaa_lumekentat(oletus, self, kentat):
  self.pyydetyt_lumekentat = getattr(
    self, 'pyydetyt_lumekentat', frozenset()
  ).union(kentat)
  # def lisaa_lumekentat


@puukota(models.sql.query.Query)
def get_select_mask(oletus, self):
  field_names, defer = self.deferred_loading
  if not field_names and defer:
    return self._get_defer_select_mask(self.get_meta(), {})
  return oletus(self)
  # def get_select_mask

@puukota(models.sql.query.Query)
def _get_defer_select_mask(oletus, self, opts, mask, select_mask=None):
  if self.model is not opts.model:
    return oletus(self, opts, mask, select_mask=select_mask)

  if select_mask is None:
    select_mask = {}

  for pyydetty_lumekentta in getattr(self, 'pyydetyt_lumekentat', ()):
    _opts, _select_mask = opts, select_mask
    for fn in pyydetty_lumekentta.split(models.sql.query.LOOKUP_SEP):
      kentta = _opts.get_field(fn)
      _select_mask = _select_mask.setdefault(kentta, {})
      try:
        _opts = kentta.remote_field.model._meta.concrete_model._meta
      except AttributeError:
        break

  def taydenna_maski(opts, mask, select_related):
    for kentta in opts.get_fields():
      if kentta.is_relation \
      and isinstance(self.select_related, dict) \
      and (_select_related := select_related.get(kentta.name)) is not None:
        if kentta.name not in mask and not taydenna_maski(
          kentta.remote_field.model._meta.concrete_model._meta,
          mask.setdefault(kentta.name, {}),
          _select_related,
        ):
          mask.pop(kentta.name)
      elif isinstance(kentta, Lumekentta) and not kentta.automaattinen:
        mask.setdefault(kentta.name, {})
    return mask
    # def taydenna_maski
  taydenna_maski(opts, mask, self.select_related)

  return oletus(self, opts, mask, select_mask=select_mask)
  # def _get_defer_select_mask


__get_deferred_fields_ohita = False
@puukota(models.Model)
def get_deferred_fields(oletus, self):
  '''
  Älä sisällytä lumekenttiä malli-olion `get_deferred_fields()`-paluuarvoon.
  Tätä joukkoa kysytään mallin tallentamisen ja kannasta lataamisen yhteydessä.
  '''
  global __get_deferred_fields_ohita
  if __get_deferred_fields_ohita:
    return oletus(self)
  return {
    kentta for kentta in oletus(self)
    if not isinstance(self._meta.get_field(kentta), Lumekentta)
  }
  # def get_deferred_fields


@puukota(models.Model)
def refresh_from_db(oletus, self, **kwargs):
  '''
  Tyhjennä mahdolliset lumekentille aiemmin lasketut arvot;
  suorita sitten tavanomainen kantakysely.
  '''
  global __get_deferred_fields_ohita
  data = self.__dict__
  for kentta in self._meta.concrete_fields:
    if isinstance(kentta, Lumekentta):
      data.pop(kentta.name, None)
      data.pop(kentta.get_attname(), None)
  if __get_deferred_fields_ohita:
    return oletus(self, **kwargs)
  __get_deferred_fields_ohita = True
  paluu = oletus(self, **kwargs)
  __get_deferred_fields_ohita = False
  return paluu
  # def refresh_from_db


@contextmanager
def _ohita_lumekentat():
  assert not hasattr(Options, 'local_fields')
  def local_fields(self):
    _local_fields = self.__dict__['local_fields']
    return type(_local_fields)(
      f for f in _local_fields
      if not isinstance(f, Lumekentta)
    )
  Options.local_fields = property(local_fields)
  try:
    yield
  finally:
    del Options.local_fields


@puukota(models.Model, koriste=classmethod)
def _check_field_name_clashes(oletus, cls):
  with _ohita_lumekentat():
    return oletus.__func__(cls)
  # def _check_field_name_clashes


@puukota(models.Model, koriste=classmethod)
def _check_model(oletus, cls):
  with _ohita_lumekentat():
    return oletus.__func__(cls)



# -*- coding: utf-8 -*-

import functools
from inspect import signature

from django.conf import settings
from django.db import models
from django.utils.functional import cached_property, classproperty

from .maare import Lumemaare, LumeFM2OMaare
from .sarake import Lumesarake


EI_ASETETTU = object()


class Lumekentta(models.fields.Field):

  # Django 5+: generoitu kenttä, joka ohitetaan tallennettaessa.
  generated = True

  @classproperty
  def forward_related_accessor_class(cls):
    # pylint: disable=no-self-argument, invalid-name, no-member
    # Huomaa, että super-toteutusta ei ole määritelty kaikille kenttätyypeille.
    # Tällaisen kentän tapauksessa kutsuvaan koodiin nousee AttributeError.
    @functools.wraps(super().forward_related_accessor_class, updated=())
    class forward_related_accessor_class(
      LumeFM2OMaare,
      super().forward_related_accessor_class
    ): pass
    return forward_related_accessor_class
    # def forward_related_accessor_class

  @classproperty
  def descriptor_class(cls):
    # pylint: disable=no-self-argument, invalid-name
    @functools.wraps(super().descriptor_class, updated=())
    class descriptor_class(Lumemaare, super().descriptor_class): pass
    return descriptor_class
    # def descriptor_class

  def __init__(
    self, *args,
    kysely, laske=None, aseta=None, automaattinen=False,
    **kwargs
  ):
    '''
    Alustaa lumekentän.
    Args:
      kysely (`django.db.models.Expression` / `lambda`): kysely
      laske (`lambda self`): paikallinen laskentafunktio
      aseta (`lambda *args`): paikallinen arvon asetusfunktio
      automaattinen (`bool`): lisätäänkö kenttä automaattisesti kyselyyn?
    '''
    # Lisää super-kutsuun parametri `editable=False`,
    # jos `aseta`-funktiota ei ole määritetty.
    if not aseta:
      kwargs['editable'] = False

    super().__init__(*args, **kwargs)

    self.default = EI_ASETETTU

    self._kysely = kysely
    self._laske = laske
    self._aseta = aseta
    self.automaattinen = automaattinen

    self.serialize = False
    # def __init__

  def deconstruct(self):
    name, path, args, kwargs = super().deconstruct()
    return name, path, args, dict(
      kysely=self.kysely,
      **kwargs
    )
    # def deconstruct

  def formfield(self, **kwargs):
    ''' Nollataan lomakkeelle annettu `initial`-arvo (EI_ASETETTU). '''
    return super().formfield(**{**kwargs, 'initial': None})

  @property
  def kysely(self):
    ''' Hae kyselylauseke (joko lambda tai suora arvo) '''
    if not callable(self._kysely):
      return self._kysely
    elif signature(self._kysely).parameters:
      return self._kysely(kentta=self)
    return self._kysely()
    # def kysely
  @kysely.setter
  def kysely(self, kysely):
    self._kysely = kysely
    # def kysely

  def laske_paikallisesti(self, rivi, select_related=False):
    '''
    Lasketaan kentän arvo paikallisesti, jos
    - laskentafunktio on määritelty; ja
    - laskentafunktio palauttaa muun arvon kuin EI_ASETETTU.

    Muuten kysytään kenttää erikseen kannasta, mikäli rivi on olemassaoleva.
    Tällöin sovelletaan `LUME_PAIKALLINEN_LASKENTA`-asetusparametriä:
    - "raise": nostetaan poikkeus
    - "print": tulostetaan tieto
    - "breakpoint": keskeytetään suoritus.

    Uudelle riville arvona palautuu `None` silloin, kun paikallista
    laskentafunktiota ei ole määritelty.

    Parametrillä `select_related=True` palautetaan kokonainen, viitattu
    (M2O tai O2O) rivi kannasta.
    '''
    # pylint: disable=protected-access
    if callable(self._laske):
      arvo = self._laske(rivi)
      if arvo is not EI_ASETETTU:
        if select_related:
          return arvo
        return getattr(arvo, 'pk', arvo)
      # if callable

    # Ohitetaan tarpeeton kysely silloin, kun rivi on uusi.
    if rivi.pk is None:
      return None

    # Kysytään erikseen kannasta.
    # Vrt. `django.db.models.Model.refresh_from_db`.
    qs = rivi.__class__._base_manager.db_manager(
      None, hints={'instance': rivi}
    ).filter(pk=rivi.pk).only(self.get_attname())
    # Ei toimi toistaiseksi.
    #if select_related:
    #  qs = qs.select_related(self.name)
    if hasattr(settings, 'CONFIG'):
      if (paikallinen := settings.CONFIG(
        'LUME_PAIKALLINEN_LASKENTA',
        default=None
      )) == 'raise':
        raise RuntimeError(
          f'Kysytään lumekenttä erikseen:'
          f' {self.model._meta.label}.{self.name}'
        )
      elif paikallinen == 'print':
        print(
          f'Kysytään lumekenttä erikseen:'
          f' {self.model._meta.label}.{self.name}'
        )
      elif paikallinen == 'breakpoint':
        breakpoint()
    try:
      #return getattr(qs.get(), self.name if select_related else self.get_attname())
      return getattr(qs.get(), self.get_attname())
    except qs.model.DoesNotExist:
      return None
    # def laske_paikallisesti

  def aseta_paikallisesti(self, rivi, arvo):
    '''
    Asetetaan kentän arvo paikallisesti, jos asetusfunktio on annettu;
    muuten nostetaan poikkeus.
    '''
    if not callable(self._aseta):
      raise RuntimeError('Asetusfunktiota ei ole määritetty: %s' % self)
    return self._aseta(rivi, arvo)
    # def aseta_paikallisesti

  def get_col(self, alias, output_field=None):
    # Ks. `Field.get_col`.
    if output_field is None:
      if isinstance(self, models.ForeignKey):
        # Ks. `ForeignKey.get_col`.
        # pylint: disable=no-member
        output_field = self.target_field
        while isinstance(output_field, models.ForeignKey):
          output_field = output_field.target_field
          if output_field is self:
            raise ValueError('Cannot resolve output_field.')
        # if isinstance
      else:
        output_field = self
    if alias != self.model._meta.db_table or output_field != self:
      return Lumesarake(alias, self, output_field)
    else:
      return self.cached_col
    # def get_col

  @cached_property
  def cached_col(self):
    return Lumesarake(self.model._meta.db_table, self)
    # def cached_col

  # Poistuu Django 6.0:ssa.
  def get_joining_columns(self, reverse_join=False):
    ''' Ohita normaali JOIN-ehto (`a`.`id` = `b`.`a_id`) '''
    # pylint: disable=unused-argument
    return tuple()
    # def get_joining_columns

  def get_joining_fields(self, reverse_join=False):
    ''' Ohita normaali JOIN-ehto (`a`.`id` = `b`.`a_id`) '''
    # pylint: disable=unused-argument
    return tuple()
    # def get_joining_columns

  def get_extra_restriction(self, alias, related_alias):
    '''
    Luo JOIN-ehto muotoa (`a`.`id` = (SELECT ... from `b`)).

    Tätä kutsutaan vain `ForeignObject`-tyyppiselle kentälle.
    '''
    # pylint: disable=unused-argument, no-member
    rhs_field = self.related_fields[0][1]
    field = rhs_field.model._meta.get_field(rhs_field.column)
    return field.get_lookup('exact')(
      self.get_col(related_alias),
      field.get_col(alias),
    )
    # def get_extra_restriction

  # class Lumekentta



# -*- coding: utf-8 -*-

from django.db import models


class Lumesarake(models.expressions.Col):
  '''
  Sarakeluokka, jonka arvo lasketaan kentälle määritetyn kyselyn mukaan.
  '''
  # pylint: disable=abstract-method

  def as_sql(self, compiler, connection):
    ''' Muodosta SELECT-lauseke ja siihen liittyvät SQL-parametrit. '''
    # pylint: disable=unused-argument
    join = compiler.query.alias_map.get(self.alias)
    if isinstance(join, models.sql.datastructures.Join):
      # Liitostaulu: muodosta alikysely tähän tauluun,
      # rajaa kysytty rivi liitostaulun primääriavaimen mukaan.
      return compiler.compile(
        models.Subquery(
          self.target.model.objects.filter(
            pk=models.expressions.RawSQL(
              '%s.%s' % (
                compiler.quote_name_unless_alias(join.table_alias),
                connection.ops.quote_name(self.target.model._meta.pk.get_attname()),
              ), ()
            ),
          ).values(**{
            # Käytetään kentän nimestä poikkeavaa aliasta.
            f'_{self.target.name}_join': self.target.kysely,
          }),
          output_field=self.field,
        ).resolve_expression(query=compiler.query)
      )
      # if isinstance(join, Join)

    elif isinstance(join, models.sql.datastructures.BaseTable):
      # Kyselyn aloitustaulu: suora kysely.
      return compiler.compile(self.target.kysely.resolve_expression(
        query=compiler.query
      ))
      # elif isinstance (join, BaseTable)

    elif join is None:
      raise NotImplementedError(f'{join!r} is None')

    else:
      # Muita kyselytyyppejä ei tueta.
      raise NotImplementedError(
        f'not isinstance({join!r}, (BaseTable, Join))'
      )
    # def as_sql

  # class Lumesarake



# -*- coding: utf-8 -*-

from django.db import models


class Lumemaare(models.query_utils.DeferredAttribute):

  def __get__(self, instance, cls=None):
    '''
    Kysyttäessä kenttää, jota ei luettu kannasta,
    haetaan sen arvo `laske_paikallisesti`-metodin avulla.

    Vrt. super.
    '''
    if instance is None:
      return self
    data = instance.__dict__
    field_name = self.field.get_attname()
    if data.get(field_name, self) is self:
      val = self._check_parent_chain(instance)
      if val is None:
        # Lasketaan arvo paikallisesti tai kysytään kannasta.
        val = self.field.laske_paikallisesti(
          instance,
          select_related=False,
        )
      data[field_name] = val
    return data[field_name]
    # def __get__

  def __set__(self, instance, value):
    '''
    Jos kentän arvo asetetaan suoraan kutsuvasta koodista,
    kutsutaan `aseta_paikallisesti`-metodia.
    '''
    if instance is None or value is self.field.default:
      return
    data = instance.__dict__
    field_name = self.field.get_attname()

    # Uusi rivi: asetetaan paikallisesti ja lisätään dataan.
    if not data.get(instance._meta.pk.get_attname()):
      self.field.aseta_paikallisesti(instance, value)
      data[field_name] = value

    # Olemassaoleva rivi: verrataan mahdolliseen jo laskettuun
    # tai aiemmin noudettuun arvoon.
    else:
      aiempi_arvo = data.get(field_name, self)
      if aiempi_arvo is self:
        aiempi_arvo = self._check_parent_chain(instance) or self
      if aiempi_arvo is self:
        data[field_name] = value
      elif aiempi_arvo != value:
        self.field.aseta_paikallisesti(instance, value)
        data[field_name] = value
      # else:
      #   Jos arvo on jo olemassa eikä se muutu, ei tehdä mitään.
      #   Muuten mm. `django.db.models.query.ModelIterable.__iter__` kaatuu.
    # def __set__

  # class Lumemaare


class LumeFM2OMaare:
  def get_object(self, instance):
    '''
    Kysyttäessä viittausta toiseen malliin, jota ei ole välimuistissa,
    haetaan sen sisältö `laske_paikallisesti(select_related=True)`-kutsun
    avulla.

    Vrt. esim. ForwardManyToOneDescriptor.get_object.
    '''
    # pylint: disable=no-member
    # Huomaa, että `self.field` asetetaan kaikille kenttätyyppikohtaisille
    # Django-kuvaajille, joista käsillä oleva luokka voidaan periyttää.
    tulos = self.field.laske_paikallisesti(
      instance,
      select_related=True,
    )
    # Palautetaan tulos sellaisenaan vain, jos se on pyydetyn mallin rivi.
    if isinstance(tulos, self.field.related_model):
      return tulos
    else:
      return super().get_object(instance)
    # def get_object
  # class LumeFM2OMaare



from django.db import models

import lume


class Asiakas(models.Model):
  nimi = models.CharField(max_length=255)
  pisin_osoite = lume.ForeignKey( # pylint: disable=no-member
    'Osoite',
    on_delete=models.DO_NOTHING,
    related_name='+',
    kysely=lambda: models.Subquery(
      Osoite.objects.filter(
        asiakas=models.OuterRef('pk'),
      ).annotate(
        osoitteen_pituus=models.Func('osoite', function='LENGTH'),
      ).order_by('-osoitteen_pituus').values('pk')[:1],
      output_field=models.IntegerField()
    ),
    null=True,
    laske=lambda rivi: next(
      (rivi[0] for rivi in sorted(
        rivi.osoitteet.values_list(
          'pk', 'osoite'
        ),
        key=lambda o: len(o[1]),
        reverse=True
      )),
      None
    ),
  )
  vanhin_lasku = lume.ForeignKey( # pylint: disable=no-member
    'Lasku',
    on_delete=models.DO_NOTHING,
    related_name='+',
    kysely=lambda: models.Subquery(
      Lasku.objects.filter(
        asiakas=models.OuterRef('pk'),
      ).order_by('numero').values('pk')[:1],
      output_field=models.IntegerField()
    ),
    laske=lambda rivi: next(
      (rivi[0] for rivi in sorted(
        rivi.laskut.values_list(
          'pk', 'numero'
        ),
        key=lambda o: o[1],
      )),
      None
    ),
    null=True,
  )
  viimeisin_lasku = lume.ForeignKey( # pylint: disable=no-member
    'Lasku',
    on_delete=models.DO_NOTHING,
    related_name='+',
    kysely=lambda: models.Subquery(
      Lasku.objects.filter(
        asiakas=models.OuterRef('pk'),
      ).order_by('-numero').values('pk')[:1],
      output_field=models.IntegerField()
    ),
    null=True,
    laske=lambda rivi: next(
      (rivi[0] for rivi in sorted(
        rivi.laskut.values_list(
          'pk', 'numero'
        ),
        key=lambda o: o[1],
        reverse=True
      )),
      None
    ),
  )
  useita_laskuja = lume.BooleanField( # pylint: disable=no-member
    kysely=~models.Q(viimeisin_lasku=models.F('vanhin_lasku')),
    laske=lambda rivi: rivi.viimeisin_lasku != rivi.vanhin_lasku,
  )

class Osoite(models.Model):
  asiakas = models.ForeignKey(
    Asiakas, related_name='osoitteet', on_delete=models.CASCADE
  )
  osoite = models.CharField(max_length=255)

class Paamies(models.Model):
  nimi = models.CharField(max_length=255)
  laskujen_summa = lume.DecimalField( # pylint: disable=no-member
    max_digits=11,
    decimal_places=2,
    kysely=lambda: models.Subquery(
      Rivi.objects.filter(
        lasku__paamies=models.OuterRef('pk'),
      ).order_by('lasku__paamies').values('lasku__paamies').values(
        summa_yht=models.Sum('summa'),
      ),
      output_field=models.DecimalField(),
    ),
  )

class Lasku(models.Model):
  asiakas = models.ForeignKey(
    Asiakas, related_name='laskut', on_delete=models.CASCADE
  )
  paamies = models.ForeignKey(Paamies, on_delete=models.CASCADE)
  numero = models.IntegerField()
  arvokkain_rivi = lume.ForeignKey( # pylint: disable=no-member
    'Rivi',
    on_delete=models.DO_NOTHING,
    related_name='+',
    kysely=lambda: models.Subquery(
      Rivi.objects.filter(
        lasku=models.OuterRef('pk'),
      ).order_by(
        '-summa'
      ).values(
        'lasku'
      ).values(
        'pk'
      ),
      output_field=models.IntegerField()
    ),
    null=True,
  )
  rivien_summa = lume.DecimalField( # pylint: disable=no-member
    max_digits=11,
    decimal_places=2,
    kysely=lambda: models.Subquery(
      Rivi.objects.filter(
        lasku=models.OuterRef('pk'),
      ).values(
        'lasku',
      ).order_by(
        'lasku'
      ).values(
        summa_yht=models.Sum('summa'),
      ),
      output_field=models.DecimalField(),
    ),
  )
  arvokkain_osuus = lume.DecimalField( # pylint: disable=no-member
    max_digits=3,
    decimal_places=2,
    kysely=models.F('arvokkain_rivi__summa') / models.F('rivien_summa'),
  )
  arvo_yli_500 = lume.BooleanField(
    kysely=models.Q(rivien_summa__gte=500),
    laske=lambda rivi: (
      rivi.rivien_summa is not None
      and rivi.rivien_summa >= 500
    ),
  )

class Rivi(models.Model):
  lasku = models.ForeignKey(Lasku, on_delete=models.CASCADE)
  summa = models.DecimalField(max_digits=11, decimal_places=2)
  selite = models.CharField(max_length=255)
